import re
from datetime import datetime, timedelta, tzinfo
from dateutil import parser as dateparser
import pytz
from typing import Any, Optional

import logging
logger = logging.getLogger(__name__)


def parse_time(timestr: str):
    return dateparser.parse(timestr).astimezone(pytz.utc).replace(tzinfo=None)


def _str(obj: Any) -> Optional[str]:
    return str(obj) if obj else None


class FixedOffset(tzinfo):
    """Fixed offset in minutes west from UTC."""

    def __init__(self, offset, name):
        self.__offset = timedelta(seconds=-offset)
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return timedelta(0)


def obj_from_dict(d):
    top = type('new', (object,), d)
    seqs = tuple, list, set, frozenset
    callables = {}
    for i, j in d.items():
        if isinstance(j, dict):
            value = obj_from_dict(j)
        elif isinstance(j, seqs):
            value = type(j)(obj_from_dict(sj) if isinstance(sj, dict) else sj for sj in j)
        else:
            value = j
        _match = re.match(r'(.*)\((.*)\)', i)
        _match2 = re.match(r'datetime\((.*)\)', j) if isinstance(j, str) else None
        if _match:
            _key = _match.groups()[0]
            # Matched parentheses, so we will need to build a function (later)
            # for _key
            _args = []
            if len(_match.groups()) > 1 and _match.groups()[1] != '':
                _args = _match.groups()[1].split(",")
            callables[_key] = callables.get(_key, [])
            # Store a list of possible arguments and their corresponding value
            callables[_key].append((_args, value))
        elif _match2:
            _datetimeval = _match2.groups()[0]
            value = datetime.fromtimestamp(int(_datetimeval))
            setattr(top, i, value)
        else:
            setattr(top, i, value)

    for _key, arg_list in callables.items():
        def _func(*args, **kwargs):
            for _data in arg_list:
                if list(args[:len(_data[0])]) == _data[0]:
                    return _data[1]
        setattr(top, _key, _func)

    return top



from sqlalchemy import Column, Integer, String, DateTime, Text, PickleType
from datetime import datetime, timedelta
from typing import Optional

from tracklater.database import db
from tracklater import settings

import logging
logger = logging.getLogger(__name__)


class ApiCall(db.Model):
    pk: int = Column(Integer, primary_key=True)
    module: str = Column(String(50), nullable=False)
    created = Column(DateTime, default=datetime.utcnow)
    start_date: datetime = Column(DateTime)
    end_date: Optional[datetime] = Column(DateTime)


class Project(db.Model):
    __tablename__ = 'projects'
    module: str = Column(String(50), primary_key=True)
    pid: str = Column(String(50), primary_key=True, nullable=True)
    group: str = Column(String(50))
    title: str = Column(String(50))

    def to_dict(self):
        return {
            "title": self.title,
            "id": self.pid,
            "group": self.group,
        }


class Issue(db.Model):
    __tablename__ = 'issues'
    module: str = Column(String(50), primary_key=True)
    id: str = Column(String(50), primary_key=True, nullable=True)
    key: str = Column(String(50), primary_key=True, nullable=True)
    group: str = Column(String(50))
    title: str = Column(String(50))
    uuid: Optional[str] = Column(String(50))
    extra_data: dict = Column(PickleType)  # For custom js

    def to_dict(self):
        return {
            "id": self.id,
            "title": self.title,
            "key": self.key,
            "group": self.group,
            "extra_data": self.extra_data,
            "uuid": self.uuid
        }


class Entry(db.Model):
    __tablename__ = 'entries'
    module: str = Column(String(50), primary_key=True)
    id: str = Column(String(50), primary_key=True, nullable=True, unique=True)
    start_time: datetime = Column(DateTime, primary_key=True)
    group: Optional[str] = Column(String(50))
    end_time: Optional[datetime] = Column(DateTime)
    date_group: Optional[str] = Column(String(50))
    issue: Optional[str] = Column(String(50))  # Issue id
    project: Optional[str] = Column(String(50))  # Project id
    title: str = Column(String(255), default="")  # Title to show in timeline
    text: str = Column(Text())  # Text to show in timeline hover
    extra_data: dict = Column(PickleType)  # For custom js

    def __init__(self, **kwargs):
        super(Entry, self).__init__(**kwargs)
        # Calculate date_group immediately
        item_time = self.start_time
        _cutoff = getattr(settings, 'CUTOFF_HOUR', 3)
        if item_time.hour >= _cutoff:
            self.date_group = item_time.strftime('%Y-%m-%d')
        else:
            self.date_group = (item_time - timedelta(days=1)).strftime('%Y-%m-%d')

    @property
    def duration(self) -> int:
        if not self.end_time:
            return 0
        return int((self.end_time - self.start_time).total_seconds())

    def to_dict(self):
        return {
            "start_time": self.start_time,
            "end_time": self.end_time,
            "id": self.id,
            "date_group": self.date_group,
            "issue": self.issue,
            "project": self.project,
            "title": self.title,
            "text": self.text,
            "extra_data": self.extra_data,
            "duration": self.duration,
            "group": self.group
        }




from slack import WebClient
from datetime import datetime
import pytz
from typing import List

from tracklater import settings
from .interfaces import EntryMixin, AbstractParser, AbstractProvider
from tracklater.models import Entry

import logging
logger = logging.getLogger(__name__)


class Parser(EntryMixin, AbstractParser):

    def get_entries(self) -> List[Entry]:
        # start_date, end_date = self.get_offset_dates()
        # if not start_date and not end_date:
        #     return []
        start_date = self.start_date
        end_date = self.end_date
        entries = []
        for group, group_data in settings.SLACK.items():

            slack_token = group_data['API_KEY']
            user_id = group_data['USER_ID']

            provider = Provider(slack_token)
            users_list = provider.api_call("users.list")
            users = {}
            for user in users_list['members']:
                users[user['id']] = (
                    user['profile'].get('first_name', 'NULL') + ' ' +
                    user['profile'].get('last_name', 'NULL')
                )
            channels = provider.api_call(
                "conversations.list", data={'types': 'public_channel,private_channel,mpim,im'}
            )['channels']
            for channel in channels:
                history = provider.api_call(
                    "conversations.history",
                    data={
                        'channel': channel['id'],
                        'oldest': (start_date - datetime(1970, 1, 1)).total_seconds(),
                        'latest': (end_date - datetime(1970, 1, 1)).total_seconds()
                    }
                )

                # Get either Istant Message recipient or channel name
                if channel.get('is_im', False) and channel.get('user', ''):
                    channel_info = users.get(channel.get('user', ''), None)
                else:
                    channel_info = channel.get('name_normalized', 'Unknown')

                for message in history['messages']:
                    if message.get('user', '') == user_id:
                        start_time = datetime.fromtimestamp(float(message['ts']))
                        # "Guess" That the timestamp has an offset equal to settings.TIMEZONE
                        if getattr(settings, 'TIMEZONE', None):
                            start_time = pytz.timezone("Europe/Helsinki").localize(
                                start_time
                            ).astimezone(pytz.utc).replace(tzinfo=None)
                        # Replace @User id with the name
                        for _user_id in users.keys():
                            if _user_id in message['text']:
                                message['text'] = message['text'].replace(
                                    _user_id, users[_user_id]
                                )
                        entries.append(Entry(
                            start_time=start_time,
                            title='',
                            text="{} - {} \n {}".format(group, channel_info, message['text']),
                            group=group
                        ))
        return entries


class Provider(AbstractProvider):
    def __init__(self, slack_token):
        self.sc = WebClient(slack_token)

    def api_call(self, *args, **kwargs):
        return self.sc.api_call(*args, **kwargs)

    def test_api_call(self, *args, **kwargs):
        if args[0] == "users.list":
            return {
                "members": [
                    {
                        "id": "1",
                        "profile": {
                            "first_name": "Firstname",
                            "last_name": "Lastename"
                        }
                    },
                    {
                        "id": "2",
                        "profile": {
                            "first_name": "Secondname",
                            "last_name": "Lastename"
                        }
                    }
                ]
            }
        if args[0] == "conversations.list":
            return {"channels": [{"id": "1"}]}
        if args[0] == "conversations.history":
            return {
                "messages": [
                    {
                        "user": "1",
                        "text": "First Message",
                        "ts": "1234567"
                    },
                    {
                        "user": "1",
                        "text": "Second Message",
                        "ts": "1234568"
                    },
                    {
                        "user": "2",
                        "text": "Third Message",
                        "ts": "1234569"
                    }
                ]
            }



from tracklater.timemodules.gitmodule import Parser

import pytest
import os
from datetime import datetime, timedelta

DIRECTORY = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture(autouse=True)
def mock_git(monkeypatch):
    monkeypatch.setattr('tracklater.timemodules.gitmodule.git_time_to_datetime',
                        lambda x: datetime.utcnow() - timedelta(days=4))


@pytest.fixture()
def parser():
    _parser = Parser(datetime.utcnow() - timedelta(days=7), datetime.utcnow())
    _parser.credentials = ('', '')
    return _parser


def test_get_entries(parser):
    """
    No real tests for gitmodule... yet.
    """
    data = parser.get_entries()
    assert len(data) == 24



from flask import Flask
import os
from tracklater.database import db
from tracklater.settings_utils import settings_wrapper as settings  # noqa

import logging
logger = logging.getLogger(__name__)


def create_app(name=__name__):
    app = Flask(name)

    DIRECTORY = os.path.dirname(os.path.realpath(__file__))

    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}/database.db'.format(DIRECTORY)
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    from tracklater.models import ApiCall, Project, Issue, Entry  # noqa

    db.init_app(app)
    with app.app_context():
        db.create_all()

    from tracklater import views

    app.register_blueprint(views.bp)

    return app


def run():
    app = create_app(name="tracklater")
    app.run(debug=True, port=5000, host="localhost")




import pytest
from tracklater import settings
from tracklater import test_settings


@pytest.fixture(autouse=True)
def mock_settings(monkeypatch):
    """
    Replace settings completely with test_settings
    """
    for module_setting in [item for item in dir(settings) if not item.startswith("__")]:
        if module_setting == 'helper':
            continue
        monkeypatch.setattr(
            'tracklater.settings.{}'.format(module_setting),
            getattr(test_settings, module_setting, {})
        )


@pytest.fixture()
def db():
    from tracklater.database import db
    return db



from typing import List
import git
import pytz
import os
import json
from datetime import datetime

from tracklater.utils import obj_from_dict
from tracklater import settings
from tracklater.timemodules.interfaces import EntryMixin, AbstractParser, AbstractProvider
from tracklater.models import Entry

import logging
logger = logging.getLogger(__name__)


def get_setting(key, default=None, group='global'):
    return settings.helper('GIT', key, group=group, default=default)


FIXTURE_DIR = os.path.dirname(os.path.realpath(__file__)) + "/fixture"


def git_time_to_datetime(_datetime):
    # Git has timezone-aware unix timestamps, convert that to a UTC datetime
    return _datetime.astimezone(pytz.utc).replace(tzinfo=None)


class Parser(EntryMixin, AbstractParser):
    def get_entries(self) -> List[Entry]:
        start_date = self.start_date
        end_date = self.end_date
        log = []
        provider = Provider()
        for group, data in settings.GIT.items():
            for repo_path in data.get('REPOS', []):
                for log_entry in provider.get_log_entries(repo_path, start_date=start_date):
                    if log_entry.author.email not in settings.GIT['global']['EMAILS']:
                        logger.info(log_entry.author.email)
                        continue
                    time = git_time_to_datetime(log_entry.authored_datetime)
                    if time < start_date or time > end_date:
                        continue

                    log.append(Entry(
                        text="{} - {}".format(repo_path.split('/')[-1], log_entry.message),
                        start_time=time,
                        group=group,
                    ))
        return log


class Provider(AbstractProvider):
    def get_log_entries(self, repo_path, start_date=None):
        repo = git.Repo(repo_path)
        for head in repo.heads:
            iterator = repo.iter_commits(head)
            for commit in iterator:
                try:
                    if start_date and git_time_to_datetime(commit.authored_datetime) < start_date:
                        break
                except Exception as e:
                    logger.warning(e)
                    continue
                yield commit

    def test_get_log_entries(self, repo_path, start_date=None):
        with open(FIXTURE_DIR + '/git_test_data.json', 'r') as f:
            _git = obj_from_dict(json.load(f))

        for commit in _git.commits:
            yield commit



import json
import os
from datetime import timedelta
from typing import List, Any, Optional

from tracklater import settings
from tracklater.utils import parse_time
from .interfaces import EntryMixin, AbstractParser, AbstractProvider
from tracklater.models import Entry

import logging
logger = logging.getLogger(__name__)


def get_setting(key, default=None, group='global') -> Any:
    return settings.helper('THYME', key, group=group, default=default)


DEFAULTS = {
    'IDLE': 900,
    'CUTOFF': 300,
}


def get_window(entry, id) -> Optional[dict]:
    for w in entry['Windows']:
        if w['ID'] == id:
            return w
    return None


class Parser(EntryMixin, AbstractParser):
    """
    Only implements "get".
    """

    def get_entries(self) -> List[Entry]:
        snapshot_entries: List[Entry] = self._read_files()
        return self._generate_sessions(snapshot_entries)

    def _read_files(self):
        provider = Provider()
        _ret: List[dict] = []
        for snapshot in provider.read_files(self.start_date, self.end_date):
            entry = self._parse_snapshot_entry(snapshot)
            if entry:
                _ret.append(entry)
        return _ret

    def _parse_snapshot_entry(self, entry):
        active_window = get_window(entry, entry['Active'])
        if active_window is None:
            return None
        return {
            'active_window': active_window,
            'time': parse_time(entry['Time']),
            'category': 'work',
        }

    def _generate_sessions(self, entries):
        def _init_session(entry):
            return Entry(
                start_time=entry['time'],
                extra_data={
                    'windows': {},
                    'group': {},
                }
            )

        def _end_session(session, entry):
            session.end_time = entry['time']
            extra = session.extra_data
            extra['windows'] = [
                {'name': window, 'time': extra['windows'][window]}
                for window in extra['windows']
            ]
            session.text = "\n".join([
                "{}s - {}".format(int(data['time']), data['name'])
                for data in sorted(extra['windows'], key=lambda x: x["time"], reverse=True)
            ])

            sorted_groups = sorted(extra['group'].items(), key=lambda val: val[1], reverse=True)

            session.extra_data['groups'] = sorted_groups
            if sorted_groups:
                session.group = sorted_groups[0][0]
                session.text = session.group + "\n" + session.text

        def _add_window(session, window, seconds):
            if 'eero@eero-ThinkPad-L470' in window['Name']:
                return session

            if window['Name'] not in session.extra_data['windows']:
                session.extra_data['windows'][window['Name']] = 0
            session.extra_data['windows'][window['Name']] += seconds

            for key in settings.THYME:
                for keyword in settings.THYME[key].get('KEYWORDS', []):
                    if keyword in window['Name']:
                        if key not in session.extra_data['group']:
                            session.extra_data['group'][key] = 0
                        session.extra_data['group'][key] += seconds

            return session

        if not entries:
            return []
        prev_entry = entries[0]
        sessions = [_init_session(prev_entry)]

        for entry in entries[1:]:
            if prev_entry['active_window'] == entry['active_window']:
                continue

            # Time spent in previous window
            diff = (entry['time'] - prev_entry['time']).total_seconds()
            session_length = (prev_entry['time'] - sessions[-1].start_time).total_seconds()

            # Add window name and time spent to extra data
            _add_window(sessions[-1], prev_entry['active_window'], diff)

            _idle = get_setting('IDLE', DEFAULTS['IDLE'])
            _cutoff = get_setting('CUTOFF', DEFAULTS['CUTOFF'])

            if (diff >= _idle or session_length >= _cutoff):
                _end_session(sessions[-1], prev_entry)
                sessions.append(_init_session(entry))

            prev_entry = entry

        _end_session(sessions[-1], prev_entry)

        return sessions


class Provider(AbstractProvider):
    def read_files(self, start_date, end_date) -> List[dict]:
        snapshot_entries = []
        filenames = []
        date = start_date
        while date <= end_date:
            filenames.append('{}/{}.json'.format(get_setting('DIR'), date.strftime('%Y-%m-%d')))
            date = date + timedelta(days=1)

        for filename in filenames:
            logging.info("opening file {}".format(filename))
            if not os.path.exists(filename):
                continue
            with open(filename) as f:
                try:
                    data = json.load(f)
                except Exception as e:
                    logger.exception(e)
                    continue
                entries = data.get('Snapshots')
                for entry in entries:
                    snapshot_entries.append(entry)
        return snapshot_entries

    def test_read_files(self, start_date=None, end_date=None):
        return [
            {
                "Time": "2019-05-23T12:00:01.242072673+03:00",
                "Windows": [
                    {
                        "ID": 1,
                        "Desktop": -1,
                        "Name": "Chrome - github"
                    },
                    {
                        "ID": 2,
                        "Desktop": 0,
                        "Name": "VSCode"
                    }
                ],
                "Active": 1,
                "Visible": [
                    1,
                    2,
                ]
            },
            {
                "Time": "2019-05-23T12:04:01.242072673+03:00",
                "Windows": [
                    {
                        "ID": 1,
                        "Desktop": -1,
                        "Name": "Chrome - github"
                    },
                    {
                        "ID": 2,
                        "Desktop": 0,
                        "Name": "VSCode"
                    }
                ],
                "Active": 2,
                "Visible": [
                    1,
                    2,
                ]
            },
            {
                "Time": "2019-05-23T12:08:01.242072673+03:00",
                "Windows": [
                    {
                        "ID": 1,
                        "Desktop": -1,
                        "Name": "Chrome - github"
                    },
                    {
                        "ID": 2,
                        "Desktop": 0,
                        "Name": "VSCode"
                    }
                ],
                "Active": 1,
                "Visible": [
                    1,
                    2,
                ]
            },
        ]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-


import importlib
from typing import Dict
from types import ModuleType

from tracklater import settings
from tracklater.timemodules.interfaces import AbstractParser
from tracklater.models import ApiCall, Entry, Issue, Project
from tracklater.database import db

import logging
logger = logging.getLogger(__name__)


def store_parser_to_database(parser, module_name, start_date, end_date):
    Entry.query.filter(
        Entry.module == module_name, Entry.start_time >= start_date,
        Entry.start_time <= end_date
    ).delete()
    for entry in parser.entries:
        entry.module = module_name
        db.session.merge(entry)
    for issue in parser.issues:
        issue.module = module_name
        db.session.merge(issue)
    for project in parser.projects:
        project.module = module_name
        db.session.merge(project)
    db.session.add(ApiCall(
        start_date=start_date,
        end_date=end_date,
        module=module_name
    ))
    db.session.commit()


def set_parser_caching_data(parser, module_name):
    apicall = ApiCall.query.filter_by(module=module_name).order_by('created').first()
    if apicall:
        parser.set_database_values(
            start_date=apicall.start_date,
            end_date=apicall.end_date,
            issue_count=Issue.query.filter_by(module=module_name).count(),
            entry_count=Entry.query.filter_by(module=module_name).count(),
            project_count=Project.query.filter_by(module=module_name).count(),
        )


class Parser(object):
    def __init__(self, start_date, end_date, modules=None) -> None:
        self.start_date = start_date
        self.end_date = end_date
        self.modules: Dict[str, AbstractParser] = {}

        for module_name in settings.ENABLED_MODULES:
            if modules and module_name not in modules:
                continue
            module: ModuleType = importlib.import_module(
                'tracklater.timemodules.{}'.format(module_name)
            )
            if getattr(module, 'Parser', None) is None:
                logger.warning('Module %s has no Parser class', module_name)
            parser = module.Parser(self.start_date, self.end_date)  # type: ignore
            self.modules[module_name] = parser

    def parse(self) -> None:
        parsers = []
        group_to_project = {project.group: project.pid for project in Project.query.all()}
        project_to_group = {project.pid: project.group for project in Project.query.all()}
        for module_name, parser in self.modules.items():
            set_parser_caching_data(parser, module_name)
            logger.warning("Parsing %s", module_name)
            parser.parse()
            parsers.append((module_name, parser))
            for entry in parser.entries:
                if not entry.project and entry.group:
                    entry.project = group_to_project.get(entry.group, None)
                if not entry.group and entry.project:
                    entry.group = project_to_group.get(str(entry.project), None)
            store_parser_to_database(self.modules[module_name], module_name,
                                     start_date=self.start_date, end_date=self.end_date)
            logger.warning("Task done %s", module_name)



import requests
import os
import json

from .interfaces import IssueMixin, AbstractParser, AbstractProvider

from tracklater import settings
from tracklater.models import Issue

from typing import Any, List

import logging
logger = logging.getLogger(__name__)


FIXTURE_DIR = os.path.dirname(os.path.realpath(__file__)) + "/fixture"


def get_setting(key, default=None, group='global') -> Any:
    return settings.helper('JIRA', key, group=group, default=default)


class Parser(IssueMixin, AbstractParser):
    def __init__(self, *args, **kwargs) -> None:
        super(Parser, self).__init__(*args, **kwargs)

    def get_issues(self) -> List[Issue]:
        issues: List[Issue] = []
        for group, group_settings in settings.JIRA.items():
            provider = Provider(get_setting('CREDENTIALS', group=group))
            issues += self.get_group_issues(provider, group, group_settings)
        return issues

    def get_group_issues(self, provider, group, group_settings) -> List[Issue]:
        cached_issues = self.caching.issue_count or 0
        # Go back a few pages to be sure
        cached_issues = max(cached_issues - provider.ISSUES_PER_PAGE * 2, 0)
        issues: List[Issue] = []
        latest_issues = provider.fetch_issues(
            project_key=group_settings['PROJECT_KEY'],
            url=group_settings['URL']
        )
        run_once = False
        while latest_issues['total'] - (cached_issues + len(issues)) > 0 or not run_once:
            run_once = True
            logger.warning(
                'Fetching issues %s to %s',
                cached_issues + len(issues), cached_issues + len(issues) + provider.ISSUES_PER_PAGE
            )
            new_issues = provider.fetch_issues(
                url=group_settings['URL'],
                project_key=group_settings['PROJECT_KEY'],
                start_from=(cached_issues + len(issues))
            )['issues']
            logger.warning(new_issues)
            for issue in new_issues:
                exists = False
                for existing_issue in issues:
                    if existing_issue.key == issue['key']:
                        exists = True
                        break
                if exists:
                    continue
                issues.append(Issue(
                    key=issue['key'],
                    title=issue['fields']['summary'],
                    group=group,
                    extra_data={'type': issue['fields']['issuetype']['name']},
                ))
        return issues


class Provider(AbstractProvider):
    def __init__(self, credentials):
        self.credentials = (credentials[0], credentials[1])
        self.ISSUES_PER_PAGE = 100 if not getattr(settings, 'TESTING', False) else 3

    def fetch_issues(self, url, project_key, start_from=None) -> dict:
        start_str = '&startAt={}'.format(start_from) if (start_from and start_from > 0) else ''
        response = requests.get(
            '{JIRA_URL}/rest/api/2/search?jql=project={JIRA_KEY}+order+by+id&fields=key,summary'
            ',issuetype&maxResults={ISSUES_PER_PAGE}{start_str}'.format(
                JIRA_URL=url, JIRA_KEY=project_key, start_str=start_str,
                ISSUES_PER_PAGE=self.ISSUES_PER_PAGE
            ), auth=self.credentials
        )
        try:
            return response.json()
        except Exception:
            logger.error(response.text)
            return {}

    def test_fetch_issues(self, url, project_key, start_from=None) -> dict:
        print(start_from)
        if start_from == 3:
            with open(FIXTURE_DIR + '/search_results_2.json', 'r') as f:
                return json.load(f)

        with open(FIXTURE_DIR + '/search_results.json', 'r') as f:
            return json.load(f)



from tracklater.timemodules.slack import Parser

import pytest
import os

from datetime import datetime, timedelta

DIRECTORY = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture()
def parser():
    _parser = Parser(datetime.utcnow() - timedelta(days=7), datetime.utcnow())
    return _parser


def test_get_entries(parser):
    """
    """
    data = parser.get_entries()
    assert len(data) == 2



from tracklater.timemodules.toggl import Parser

import pytest
import os

from datetime import datetime, timedelta
from tracklater.models import Entry

DIRECTORY = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture()
def parser():
    _parser = Parser(datetime.utcnow() - timedelta(days=7), datetime.utcnow())
    return _parser


def test_toggl_get_entries(parser):
    """
    """
    data = parser.get_entries()
    assert len(data) == 3


def test_toggl_get_projects(parser):
    """
    """
    data = parser.get_projects()
    assert len(data) == 4


def test_toggl_add_modify_delete(parser: Parser):
    parser.entries = parser.get_entries()
    entry = Entry(
        id="4",
        start_time=datetime.utcnow() - timedelta(hours=2),
        end_time=datetime.utcnow() - timedelta(hours=1),
        title="Toggl new entry (4)",
        project="10",
    )
    entry = parser.create_entry(entry, None)
    assert entry.id == "4"

    entry.title = "Toggl modified entry"
    entry = parser.update_entry("4", entry, None)
    assert entry.title == "Toggl modified entry"

    parser.delete_entry("4")



from tracklater.timemodules.jira import Parser, Provider

import pytest
import os

from datetime import datetime, timedelta

TEST_URL = 'mock://jira.test'
TEST_KEY = 'TEST'

DIRECTORY = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture()
def parser():
    _parser = Parser(datetime.utcnow() - timedelta(days=7), datetime.utcnow())
    _parser.credentials = ('', '')
    return _parser


@pytest.fixture()
def provider():
    _provider = Provider(('', ''))
    return _provider


def test_create_parser(parser: Parser):
    assert parser is not None


def test_fetch_issues(provider: Provider):
    data = provider.fetch_issues(TEST_URL, TEST_KEY)
    assert data['issues'][0]['key'] == 'TEST-1'

    data = provider.fetch_issues(TEST_URL, TEST_KEY, start_from=3)
    assert data['issues'][0]['key'] == 'TEST-4'


def test_get_group_issues(parser: Parser, provider: Provider):
    from tracklater import settings
    issues = parser.get_group_issues(provider, 'group1', settings.JIRA['group1'])
    assert len(issues) == 6


def test_get_issues(parser: Parser):
    issues = parser.get_issues()
    assert len(issues) == 6


def test_cache(parser: Parser):
    parser.get_issues()
    issues = parser.get_issues()
    assert len(issues) == 6



from typing import List, Optional
from datetime import datetime
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

from tracklater import settings
from tracklater.models import Entry, Issue, Project


@dataclass
class CachingData:
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    issue_count: Optional[int] = None
    entry_count: Optional[int] = None
    project_count: Optional[int] = None


def testing_decorator(func):
    def _func(*args, **kwargs):
        self = args[0]

        if not getattr(settings, 'TESTING', False):
            return func(*args, **kwargs)

        if func.__name__.startswith("test_") or func.__name__.startswith("__"):
            return func(*args, **kwargs)

        if not hasattr(self, "test_{}".format(func.__name__)):
            raise NotImplementedError("No test method for {}.{}".format(self, func.__name__))

        return getattr(self, "test_{}".format(func.__name__))(*args[1:], **kwargs)

    return _func


class ProviderMetaclass(type):
    def __new__(cls, name, bases, local):
        for attr in local:
            value = local[attr]
            if callable(value):
                local[attr] = testing_decorator(value)
        return type.__new__(cls, name, bases, local)


class AbstractProvider(metaclass=ProviderMetaclass):
    """
    Will run methods prefixed "test_" when TESTING == True
    """
    pass


class AbstractParser(metaclass=ABCMeta):
    def __init__(self, start_date: datetime, end_date: datetime) -> None:
        self.start_date: datetime = start_date
        self.end_date: datetime = end_date
        self.entries: List[Entry] = []
        self.projects: List[Project] = []
        self.issues: List[Issue] = []
        self.caching: CachingData = CachingData()

    def set_database_values(self, start_date=None, end_date=None, issue_count=None,
                            entry_count=None, project_count=None) -> None:
        self.caching.start_date = start_date
        self.caching.end_date = end_date
        self.caching.issue_count = issue_count
        self.caching.entry_count = entry_count
        self.caching.project_count = project_count

    def get_offset_dates(self):
        """
        Use cached api call start and end date to get a smart timeframe to use
        e.g. We already have an api call for (Tue-Fri), and we try to get data for
        (Mon-Wed). In this case this method returns (Mon-Tue).
        """
        if not self.caching.start_date or not self.caching.end_date:
            return (self.start_date, self.end_date)

        # ---a---c---a---c------
        #    (   )
        # ---a---c-------x------
        #    (   )
        # -a---a-c-------c------
        #  (     )
        if (self.caching.start_date > self.start_date
                and self.caching.end_date >= self.end_date):
            return (self.start_date, self.caching.start_date)

        # ------c-a--a---c------
        #
        # ------x--------x------
        #
        if (self.caching.start_date <= self.start_date
                and self.caching.end_date >= self.end_date):
            return (None, None)

        # ------c---a----c--a---
        #                (  )
        # ------c--------c-a--a-
        #                (    )
        # ------x--------c---a-
        #                (   )
        if (self.caching.start_date <= self.start_date
                and self.caching.end_date < self.end_date):
            return (self.caching.end_date, self.end_date)

        # Other cases, just skip caching
        return (self.start_date, self.end_date)

    @property
    def capabilities(self) -> List[str]:
        return []

    @abstractmethod
    def parse(self) -> None:
        pass


class EntryMixin(AbstractParser):
    def parse(self) -> None:
        self.entries = self.get_entries()
        super(EntryMixin, self).parse()

    @property
    def capabilities(self) -> List[str]:
        _ret = super(EntryMixin, self).capabilities
        return _ret + ['entries']

    @abstractmethod
    def get_entries(self) -> List[Entry]:
        raise NotImplementedError()


class ProjectMixin(AbstractParser):
    def parse(self) -> None:
        self.projects = self.get_projects()
        super(ProjectMixin, self).parse()

    @property
    def capabilities(self) -> List[str]:
        _ret = super(ProjectMixin, self).capabilities
        return _ret + ['projects']

    @abstractmethod
    def get_projects(self) -> List[Project]:
        raise NotImplementedError()


class IssueMixin(AbstractParser):
    def parse(self) -> None:
        self.issues = self.get_issues()
        super(IssueMixin, self).parse()

    @property
    def capabilities(self) -> List[str]:
        _ret = super(IssueMixin, self).capabilities
        return _ret + ['issues']

    @abstractmethod
    def get_issues(self) -> List[Issue]:
        raise NotImplementedError()

    def find_issue(self, uuid) -> Optional[Issue]:
        for issue in self.issues:
            if issue.uuid == uuid:
                return issue
        return None


class AddEntryMixin(AbstractParser):
    @abstractmethod
    def create_entry(self, new_entry: Entry, issue: Issue) -> Entry:
        raise NotImplementedError()

    @property
    def capabilities(self) -> List[str]:
        _ret = super(AddEntryMixin, self).capabilities
        return _ret + ['addentry']


class DeleteEntryMixin(AbstractParser):
    @abstractmethod
    def delete_entry(self, entry_id: str) -> None:
        raise NotImplementedError()

    @property
    def capabilities(self) -> List[str]:
        _ret = super(DeleteEntryMixin, self).capabilities
        return _ret + ['deleteentry']


class UpdateEntryMixin(AbstractParser):
    @abstractmethod
    def update_entry(self, entry_id: str, new_entry: Entry, issue: Issue) -> Entry:
        raise NotImplementedError()

    @property
    def capabilities(self) -> List[str]:
        _ret = super(UpdateEntryMixin, self).capabilities
        return _ret + ['updateentry']



import pytest
import os

from typing import Any
from datetime import datetime, timedelta

from tracklater import create_app
from tracklater.models import Entry, Issue, Project

DIRECTORY = os.path.dirname(os.path.realpath(__file__))

app = create_app()


@pytest.fixture
def client(db):
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}/database_testing.db'.format(DIRECTORY)
    client = app.test_client()
    db.init_app(app)
    with app.app_context():
        db.create_all()

    yield client
    os.remove("{}/database_testing.db".format(DIRECTORY))


def test_app_root(client):
    response = client.get('/')
    assert response.status_code == 200


def test_app_listmodules(client):
    response = client.get('/listmodules')
    assert response.status_code == 200


def test_app_fetchdata(client):
    response = client.get('/fetchdata')
    assert response.status_code == 200


def test_database_entries(client, db: Any):
    from tracklater.timemodules.toggl import Parser
    from tracklater.main import store_parser_to_database
    start_date = datetime.utcnow() - timedelta(days=7)
    end_date = datetime.utcnow()
    parser = Parser(start_date, end_date)
    parser.entries = parser.get_entries()
    with app.app_context():
        store_parser_to_database(parser, 'toggl', start_date, end_date)

        db_entries = Entry.query.all()
        assert len(parser.entries) == len(db_entries)


def test_database_projects(client, db: Any):
    from tracklater.timemodules.toggl import Parser
    from tracklater.main import store_parser_to_database
    start_date = datetime.utcnow() - timedelta(days=7)
    end_date = datetime.utcnow()
    parser = Parser(start_date, end_date)
    parser.projects = parser.get_projects()
    with app.app_context():
        store_parser_to_database(parser, 'toggl', start_date, end_date)

        db_projects = Project.query.all()
        assert len(parser.projects) == len(db_projects)


def test_database_issues(client, db: Any):
    from tracklater.timemodules.jira import Parser
    from tracklater.main import store_parser_to_database
    start_date = datetime.utcnow() - timedelta(days=7)
    end_date = datetime.utcnow()
    parser = Parser(start_date, end_date)
    parser.issues = parser.get_issues()
    with app.app_context():
        store_parser_to_database(parser, 'jira', start_date, end_date)

        db_issues = Issue.query.all()
        assert len(parser.issues) == len(db_issues)


@pytest.mark.skip('Broken from commit 27b780c7d23261669e8e5a997403c6a80d6bbaeb')
def test_database_jira_caching(client, db: Any):
    from tracklater.timemodules.jira import Parser
    from tracklater.main import store_parser_to_database, set_parser_caching_data
    start_date = datetime.utcnow() - timedelta(days=7)
    end_date = datetime.utcnow()
    parser = Parser(start_date, end_date)
    parser.issues = parser.get_issues()
    with app.app_context():
        store_parser_to_database(parser, 'jira', start_date, end_date)

    parser = Parser(start_date, end_date)
    # Skip caching first
    parser.issues = parser.get_issues()
    # We get 6 issues again
    assert len(parser.issues) == 6

    parser = Parser(start_date, end_date)
    with app.app_context():
        set_parser_caching_data(parser, 'jira')  # Fetch caching values
    parser.issues = parser.get_issues()
    # Since all 6 are already in database, get_issues should just run once (check jira.py)
    # Hence, wwith testing data we get the 3 last issues as a response
    assert len(parser.issues) == 3


@pytest.mark.skip('Broken from commit be24bce7a5c3e472c49c1a9e5712712501453ba5')
def test_database_slack_caching(client, db: Any):
    from tracklater.timemodules.slack import Parser
    from tracklater.main import store_parser_to_database, set_parser_caching_data
    start_date = datetime.utcnow() - timedelta(days=7)
    end_date = datetime.utcnow()
    parser = Parser(start_date, end_date)
    parser.entries = parser.get_entries()
    with app.app_context():
        store_parser_to_database(parser, 'slack', start_date, end_date)

    parser = Parser(start_date, end_date)
    # Skip caching first
    parser.entries = parser.get_entries()
    # We get 2 entries again
    assert len(parser.entries) == 2

    parser = Parser(start_date, end_date)
    with app.app_context():
        set_parser_caching_data(parser, 'slack')  # Fetch caching values
    parser.entries = parser.get_entries()
    # Since we already fetched entries for there dates we don't get any new ones
    assert len(parser.entries) == 0

    parser = Parser(start_date, end_date + timedelta(hours=1))
    with app.app_context():
        set_parser_caching_data(parser, 'slack')  # Fetch caching values
    parser.entries = parser.get_entries()
    assert len(parser.entries) == 2




TESTING = True

ENABLED_MODULES = ['jira', 'gitmodule', 'slack', 'taiga', 'toggl', 'thyme']

from datetime import datetime, timedelta

OVERRIDE_START = datetime.fromtimestamp(
    1234560
) - timedelta(days=4)
OVERRIDE_END = datetime.fromtimestamp(
    1234560
) + timedelta(days=4)

UI_SETTINGS = {
    'toggl': {
        'global': '#E01A22'
    },
    'thyme': {
        'global': '#1aef65'
    },
    'gitmodule': {
        'global': '#F44D27'
    },
    'slack': {
        'global': '#4A154B'
    }
}

# Test settings for Jira
JIRA = {
    'group1': {
        'CREDENTIALS': ('', ''),
        'URL': 'mock://jira.test',
        'PROJECT_KEY': 'TEST',
    }
}

# Test settings for Git
GIT = {
    'global': {
        'EMAILS': ['test.person@email.com'],
    },
    'group1': {
        'REPOS': ['path1', 'path2']
    },
    'group2': {
        'REPOS': ['path3']
    },
}

# Test settings for Slack
SLACK = {
    'global': {
        'API_KEY': '',
        'USER_ID': '1',
    }
}

# Test settings for Slack
TAIGA = {
    'global': {
        'CREDENTIALS': 'test'
    },
    'group2': {
        'project_slug': 'test'
    }
}

TOGGL = {
    'global': {
        'API_KEY': 'your-api-key'
    },
    'group1': {
        'NAME': 'First Client',
        'PROJECTS': {
            'Development': 'default',
            'Bug fixing': 'bug',
        }
    },
    'group2': {
        'NAME': 'Second Client',
        'PROJECTS': {
            'Development': 'default',
            'Bug fixing': 'default',
        }
    }
}

THYME = {  # type: ignore
    'global': {
    }
}



from flask import request, Blueprint
from tracklater.utils import _str
from datetime import datetime, timedelta, date
import json
import pytz
from typing import Dict, Any

from tracklater.database import db
from tracklater.main import Parser
from tracklater import settings
from tracklater.models import Entry, Issue, Project, ApiCall  # noqa
from tracklater.timemodules.interfaces import AddEntryMixin, UpdateEntryMixin

import logging
logger = logging.getLogger(__name__)


bp = Blueprint("main", __name__, static_folder="static", static_url_path="/static/views")


@bp.route('/', methods=['GET'])
def index():
    return bp.send_static_file('index.html')


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        if obj.tzinfo and obj.tzinfo is not pytz.utc:
            logger.warning("Trying to serialize timezone aware date %s", obj)
        obj.replace(tzinfo=None)
        # No idea, but for some reason using isoformat() here did *not* work.
        # Just add utc timezone manually then... (This will make js convert it automatically)
        return obj.isoformat() + "+00:00"
    raise TypeError("Type %s not serializable" % type(obj))


class MyEncoder(json.JSONEncoder):
    def default(self, o):
        return {k.lstrip('_'): v for k, v in vars(o).items()}


@bp.route('/listmodules', methods=['GET'])
def listmodules() -> Any:
    if request.method == 'GET':
        data = {}
        parser = Parser(None, None)
        for module_name in settings.ENABLED_MODULES:
            data[module_name] = {
                'color': settings.UI_SETTINGS.get(module_name, {}),
                'capabilities': parser.modules[module_name].capabilities,
            }
        return json.dumps(data, default=json_serial)
    return None


@bp.route('/getsettings', methods=['GET'])
def getsettings() -> Any:
    return json.dumps(settings, cls=MyEncoder)


@bp.route('/fetchdata', methods=['GET'])
def fetchdata() -> Any:
    if request.method == 'GET':
        keys = request.values.getlist('keys[]')
        parse = request.values.get('parse', '1')
        now = datetime.utcnow()
        if 'from' in request.values:
            from_date = parseTimestamp(request.values['from'])
        else:
            from_date = now - timedelta(days=41)
        if 'to' in request.values:
            to_date = parseTimestamp(request.values['to'])
        else:
            to_date = now

        if getattr(settings, 'OVERRIDE_START', None):
            from_date = settings.OVERRIDE_START
        if getattr(settings, 'OVERRIDE_END', None):
            to_date = settings.OVERRIDE_END

        if keys and keys[0] == "all":
            keys = None
        parser = Parser(from_date, to_date, modules=keys)
        if parse == '1':
            parser.parse()
        data: Dict[str, Dict] = {}
        for key in settings.ENABLED_MODULES:
            if not keys or key in keys:
                data[key] = {}
                data[key]['entries'] = [entry.to_dict()
                                        for entry in Entry.query.filter(
                                            Entry.module == key,
                                            Entry.start_time >= from_date,
                                            Entry.start_time <= to_date
                                        )]
                data[key]['projects'] = [project.to_dict()
                                         for project in Project.query.filter(
                                            Project.module == key
                                        )]
                data[key]['issues'] = [issue.to_dict()
                                       for issue in Issue.query.filter(
                                            Issue.module == key
                                        )]
                data[key]['capabilities'] = parser.modules[key].capabilities
                data[key]['color'] = settings.UI_SETTINGS.get(key, {})
        return json.dumps(data, default=json_serial)
    return None


def parseTimestamp(stamp):
    if not stamp:
        return None
    date = datetime.fromtimestamp(int(stamp) / 1e3)
    return date


@bp.route('/updateentry', methods=['POST'])
def updateentry() -> Any:
    if request.method == 'POST':
        data = request.get_json()
        module = data.get('module')
        entry_id = _str(data.get('entry_id', None))
        project = data.get('project_id', None)
        if project == "null":
            project = None
        project_to_group = {project.pid: project.group for project in Project.query.all()}
        new_entry = Entry(
            start_time=parseTimestamp(data['start_time']),
            end_time=parseTimestamp(data.get('end_time', None)),
            id=entry_id,
            issue=data.get('issue_id', None),
            project=project,
            title=data.get('title', ''),
            text=data.get('text', ""),
            extra_data=data.get('extra_data', {}),
            group=project_to_group.get(str(project), None)
        )
        issue = None
        if new_entry.issue:
            issue = Issue.query.filter(Issue.uuid == new_entry.issue).first()

        parser = Parser(None, None)

        if not entry_id:
            # Check that create is allowed
            assert isinstance(parser.modules[module], AddEntryMixin)
            new_entry = parser.modules[module].create_entry(  # type: ignore
                new_entry=new_entry,
                issue=issue
            )
        else:
            # Check that update is allowed
            assert isinstance(parser.modules[module], UpdateEntryMixin)
            new_entry = parser.modules[module].update_entry(  # type: ignore
                entry_id=new_entry.id,
                new_entry=new_entry,
                issue=issue
            )
        data = "error"

        if new_entry:
            Entry.query.filter(Entry.id == new_entry.id).delete()
            new_entry.module = module
            db.session.merge(new_entry)
            db.session.commit()
            data = new_entry.to_dict()

        return json.dumps(data, default=json_serial)
    return None


@bp.route('/deleteentry', methods=['POST'])
def deleteentry() -> Any:
    if request.method == 'POST':
        data = request.get_json()
        module = data.get('module')
        entry_id = data.get('entry_id')

        parser = Parser(None, None)
        # Check that delete is allowed
        assert isinstance(parser.modules[module], AddEntryMixin)
        ret = parser.modules[module].delete_entry(  # type: ignore
            entry_id=entry_id
        )

        Entry.query.filter(Entry.id == entry_id).delete()
        db.session.commit()

        return json.dumps(ret, default=json_serial)
    return None



import requests
from typing import List, cast, Any

from tracklater import settings
from .interfaces import IssueMixin, AbstractParser, AbstractProvider
from tracklater.models import Issue

import logging
logger = logging.getLogger(__name__)

AUTH_URL = 'https://api.taiga.io/api/v1/auth'
ISSUE_URL = 'https://api.taiga.io/api/v1/userstories'
PROJECT_URL = 'https://api.taiga.io/api/v1/projects/by_slug?slug={}'


class Parser(IssueMixin, AbstractParser):
    def get_issues(self) -> List[Issue]:
        self.taiga_login()
        issues: List[Issue] = []

        latest_issues = self.taiga_fetch_issues()
        for issue in latest_issues:
            taiga_project = [p for p in self.taiga_projects if p['id'] == issue['project']][0]
            issues.append(Issue(
                key="#{}".format(issue['ref']),
                id=issue['id'],
                title=issue['subject'],
                group=taiga_project['group']
            ))
        return issues

    def taiga_login(self) -> None:
        taiga_settings = cast(Any, settings.TAIGA)
        self.provider = Provider(taiga_settings['global']['CREDENTIALS'])
        self.taiga_projects: List[dict] = []
        # Get taiga project id for all clients
        # "No client" not supported yet
        for group, data in taiga_settings.items():
            if 'project_slug' not in data:
                continue
            project = self.provider.get_project(data['project_slug'])
            self.taiga_projects.append({
                'id': project['id'],
                'group': group
            })

    def taiga_fetch_issues(self, start_from=None):
        issues: List[dict] = []
        for taiga_project in self.taiga_projects:
            issues += self.provider.get_issues(taiga_project['id'])
        return issues


class Provider(AbstractProvider):
    def __init__(self, credentials):
        self.token = self.login(credentials)
        self.headers = {
            "Authorization": "Bearer {}".format(self.token), 'x-disable-pagination': 'True'
        }

    def login(self, credentials):
        response = requests.post(
            AUTH_URL, data={
                'type': 'normal',
                'username': credentials[0],
                'password': credentials[1]
            }
        )
        return response.json()['auth_token']

    def test_login(self, credentials):
        return ""

    def get_project(self, project_slug):
        response = requests.get(
            PROJECT_URL.format(project_slug),
            headers=self.headers
        )
        return response.json()

    def test_get_project(self, credentials):
        return {'id': "1"}

    def get_issues(self, project_id):
        response = requests.get(
            '{ISSUE_URL}?project={id}'.format(
                ISSUE_URL=ISSUE_URL,
                id=project_id
            ), headers=self.headers
        )
        return response.json()

    def test_get_issues(self, credentials):
        return [
            {
                "ref": "1",
                "id": "1",
                "subject": "Taiga issue 1",
                "project": "1"
            },
            {
                "ref": 2,
                "id": "2",
                "subject": "Taiga issue 2",
                "project": "1"
            },
            {
                "ref": 3,
                "id": "3",
                "subject": "Taiga issue 3",
                "project": "1"
            }
        ]



from tracklater.timemodules.interfaces import AbstractProvider

from tracklater.utils import obj_from_dict


def test_obj_from_dict():
    _dict = {
        'test(arg1,arg2)': {
            'foo': {
                'bar': 'The Crazy End'
            }
        },
        'test(arg1)': {
            'foo': {
                'bar': 'The Good End'
            }
        },
        'test()': {
            'foo': {
                'bar': 'The Best End'
            }
        }
    }
    obj = obj_from_dict(_dict)

    assert obj.test('arg1', 'arg2').foo.bar == 'The Crazy End'
    assert obj.test('arg1').foo.bar == 'The Good End'
    assert obj.test().foo.bar == 'The Best End'


class TestProvider(AbstractProvider):
    def normal_method(self, arg1=None):
        return "normal"

    def test_normal_method(self, arg1=None):
        return "test"


def test_provider_normal(monkeypatch):
    monkeypatch.setattr('tracklater.settings.TESTING', False)
    provider = TestProvider()

    assert provider.normal_method() == "normal"


def test_provider_testing(monkeypatch):
    provider = TestProvider()

    assert provider.normal_method() == "test"



import json
import appdirs
import os
import shutil
from typing import Any

import logging
logger = logging.getLogger(__name__)

DIRECTORY = os.path.dirname(os.path.realpath(__file__))
user_config_path = os.path.join(appdirs.user_config_dir(), 'tracklater.json')


class Dummy(object):
    pass


settings_wrapper: Any = Dummy()

if not os.path.exists(user_config_path):
    try:
        os.mkdir(appdirs.user_config_dir())
    except FileExistsError:
        pass
    shutil.copy(os.path.join(DIRECTORY, 'example_settings.json'), user_config_path)
    logger.error("No user settings file! Modify the example settings created (path: %s)",
                 user_config_path)

try:
    with open(user_config_path, 'rb') as f:
        for key, value in json.load(f).items():
            setattr(settings_wrapper, key, value)
except json.JSONDecodeError as e:
    logger.exception("Error reading settings file: %s", e)


def helper(module, key, group='global', default=None):
    # Try given group first
    if group in getattr(settings_wrapper, module):
        _tmp_dict = getattr(settings_wrapper, module)[group]
        if key in _tmp_dict:
            return _tmp_dict[key]
    # Try with 'global' group
    if 'global' in getattr(settings_wrapper, module):
        _tmp_dict = getattr(settings_wrapper, module)['global']
        if key in _tmp_dict:
            return _tmp_dict[key]
    if default:
        return default

    raise KeyError('No setting "{}" for module {} found'.format(key, module))


settings_wrapper.helper = helper



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='tracklater',
    version='1.6.1',
    description=('TrackLater helps you track time after-the-fact by combining clues and showing'
                 'your day on a simple timeline view.'),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Eero Vilpponen',
    author_email='eero.vilpponen@gmail.com',
    packages=find_packages(),
    py_modules=['tracklater'],
    install_requires=requirements,
    python_requires='>=3.7.1',
    include_package_data=True,
    entry_points={
        'console_scripts': ['tracklater = tracklater.__init__:run']
    },
    url="https://github.com/Eerovil/TrackLater",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import requests
import json
from typing import List, Union, cast, Any, Optional
from datetime import timedelta

from tracklater.utils import parse_time, _str
from tracklater import settings
from .interfaces import (
    EntryMixin, AddEntryMixin, UpdateEntryMixin, DeleteEntryMixin, ProjectMixin, AbstractParser,
    AbstractProvider
)
from tracklater.models import Entry, Project, Issue


import logging
logger = logging.getLogger(__name__)


def get_setting(key, default=None, group='global'):
    return settings.helper('CLOCKIFY', key, group=group, default=default)


class Parser(EntryMixin, AddEntryMixin, UpdateEntryMixin, DeleteEntryMixin, ProjectMixin,
             AbstractParser):
    def __init__(self, *args, **kwargs):
        super(Parser, self).__init__(*args, **kwargs)
        self.provider = Provider(get_setting('API_KEY'))
        workspaces = self.provider.request(
            'workspaces', method='GET'
        )
        if len(workspaces) > 1:
            if 'WORKSPACE' in settings.CLOCKIFY['global']:
                for workspace in workspaces:
                    if workspace['id'] == settings.CLOCKIFY['global']['WORKSPACE']:
                        self.provider.workspace = workspace
            else:
                self.provider.workspace = workspaces[0]
                logger.warning(
                    "More than one clockify workspace... Using just one (%s)",
                    self.provider.workspace['id']
                )
            
        self.user = self.provider.request(
            'user', method='GET'
        )


    def get_entries(self) -> List[Entry]:

        if self.start_date:
            params = {'start': self.start_date.isoformat() + 'Z',
                      'end': self.end_date.isoformat() + 'Z'}
        else:
            params = {}

        time_entries = []
        data = self.provider.workspace_request(
            'user/{user_id}/time-entries'.format(
                user_id=self.user['id']
            ), params=params, method='GET'
        )
        for entry in data:
            time_entries.append(Entry(
                id=entry['id'],
                start_time=parse_time(entry['timeInterval']['start']),
                end_time=parse_time(entry['timeInterval']['end']),
                title=entry['description'],
                project=entry['projectId'],
            ))

        return time_entries

    def get_projects(self) -> List[Project]:
        projects = []
        clients = self.provider.workspace_request('clients', method='GET')
        _settings = cast(Any, settings.CLOCKIFY)
        for client in clients:
            group = None
            for project, data in _settings.items():
                if data.get('NAME', None) == client['name']:
                    group = project
            if not group:
                continue
            resp = self.provider.workspace_request(
                'projects?clients={}'.format(client['id']), method='GET'
            )
            for project in resp:
                if project['name'] not in _settings[group]['PROJECTS']:
                    continue
                projects.append(Project(
                    pid=project['id'],
                    title="{} - {}".format(client['name'], project['name']),
                    group=group
                ))
        return projects

    def create_entry(self, new_entry: Entry, issue: Optional[Issue]) -> Entry:
        if not new_entry.project or new_entry.project == '0':
            new_entry.project = None
        if new_entry.end_time is None:
            raise ValueError("No end_time")
        entry = self.provider.workspace_request(
            'time-entries',
            data=json.dumps({
                'start': new_entry.start_time.isoformat() + 'Z',
                'end': new_entry.end_time.isoformat() + 'Z',
                'description': new_entry.title,
                'projectId': new_entry.project,
            }),
            method='POST'
        )
        return Entry(
            id=_str(entry['id']),
            start_time=parse_time(entry['timeInterval']['start']),
            end_time=parse_time(entry['timeInterval']['end']),
            title=entry['description'],
            project=entry['projectId'],
        )

    def update_entry(self, entry_id: str, new_entry: Entry, issue: Optional[Issue]) -> Entry:
        if not new_entry.project or new_entry.project == '0':
            new_entry.project = None
        if new_entry.end_time is None:
            raise ValueError("No end_time")
        entry = self.provider.workspace_request(
            'time-entries/{}'.format(entry_id),
            data=json.dumps({
                'start': new_entry.start_time.isoformat() + 'Z',
                'end': new_entry.end_time.isoformat() + 'Z',
                'description': new_entry.title,
                'projectId': new_entry.project,
            }),
            method='PUT'
        )
        return Entry(
            id=_str(entry['id']),
            start_time=parse_time(entry['timeInterval']['start']),
            end_time=parse_time(entry['timeInterval']['end']),
            title=entry['description'],
            project=entry['projectId'],
        )

    def delete_entry(self, entry_id: str) -> None:
        self.provider.workspace_request(
            'time-entries/{}'.format(entry_id),
            data=json.dumps({}),
            method="DELETE"
        )


class Provider(AbstractProvider):
    def __init__(self, api_key):
        self.api_key = api_key
        self.id_counter = 4
        self.workspace = None

    def workspace_request(self, endpoint: str, **kwargs) -> Union[List[dict], dict]:
        return self.request('workspaces/{workspace_id}/{endpoint}'.format(
            workspace_id=self.workspace['id'],
            endpoint=endpoint
        ), **kwargs)

    def request(self, endpoint: str, **kwargs) -> Union[List[dict], dict]:
        url = 'https://api.clockify.me/api/v1/{}'.format(endpoint)
        kwargs['headers'] = kwargs.get('headers', {
            "Content-Type": "application/json"
        })
        kwargs['headers']['X-Api-Key'] = self.api_key

        method = kwargs.get('method', 'POST').lower()
        try:
            del kwargs['method']
        except KeyError:
            pass
        response = getattr(requests, method)(url, **kwargs)
        if method == 'delete':
            return {}
        try:
            ret = response.json()
            return ret
        except Exception as e:
            logger.exception("%s: %s", response.content, e)
            raise

    def test_request(self, endpoint: str, **kwargs) -> Union[List[dict], dict, str]:
        method = kwargs.get('method', 'POST').lower()
        return [{}]



import requests
import json
import time
from typing import List, Union, cast, Any, Optional
from datetime import timedelta

from tracklater.utils import parse_time, _str
from tracklater import settings
from .interfaces import (
    EntryMixin, AddEntryMixin, UpdateEntryMixin, DeleteEntryMixin, ProjectMixin, AbstractParser,
    AbstractProvider
)
from tracklater.models import Entry, Project, Issue


import logging
logger = logging.getLogger(__name__)


def get_setting(key, default=None, group='global'):
    return settings.helper('TOGGL', key, group=group, default=default)


class Parser(EntryMixin, AddEntryMixin, UpdateEntryMixin, DeleteEntryMixin, ProjectMixin,
             AbstractParser):
    def __init__(self, *args, **kwargs):
        super(Parser, self).__init__(*args, **kwargs)
        self.provider = Provider(get_setting('API_KEY'))

    def get_entries(self) -> List[Entry]:
        if self.start_date:
            params = {'start_date': self.start_date.isoformat() + "+00:00",
                      'end_date': self.end_date.isoformat() + "+00:00"}
        else:
            params = {}
        data = self.provider.request(
            'time_entries', params=params, method='GET'
        )
        time_entries = []
        for entry in data:
            time_entries.append(Entry(
                id=entry['id'],
                start_time=parse_time(entry['start']),
                end_time=parse_time(entry['stop']),
                title=entry.get('description', ''),
                project=entry.get('pid', None),
            ))

        return time_entries

    def get_projects(self) -> List[Project]:
        clients = self.provider.request('clients', method='GET')
        projects = []
        toggl_settings = cast(Any, settings.TOGGL)
        for client in clients:
            group = None
            for project, data in toggl_settings.items():
                if data.get('NAME', None) == client['name']:
                    group = project
            if not group:
                continue
            resp = self.provider.request(
                'clients/{}/projects'.format(client['id']), method='GET'
            )
            for project in resp:
                if project['name'] not in toggl_settings[group]['PROJECTS']:
                    continue
                projects.append(Project(
                    pid=project['id'],
                    title="{} - {}".format(client['name'], project['name']),
                    group=group
                ))
        return projects

    def create_entry(self, new_entry: Entry, issue: Optional[Issue]) -> Entry:
        entry = self.push_session(
            session={
                'start_time': new_entry.start_time,
                'end_time': new_entry.end_time,
            },
            name=new_entry.title,
            project_id=new_entry.project
        )
        return Entry(
            id=_str(entry['id']),
            start_time=parse_time(entry['start']),
            end_time=parse_time(entry['stop']),
            title=entry['description'],
            project=entry.get('pid', None),
            group=new_entry.group,
        )

    def update_entry(self, entry_id: str, new_entry: Entry, issue: Optional[Issue]) -> Entry:
        updated_entry = self.push_session(
            session={
                'start_time': new_entry.start_time,
                'end_time': new_entry.end_time,
            },
            entry_id=entry_id,
            name=new_entry.title,
            project_id=new_entry.project
        )

        return Entry(
            id=_str(updated_entry['id']),
            start_time=parse_time(updated_entry['start']),
            end_time=parse_time(updated_entry['stop']),
            title=updated_entry['description'],
            project=_str(updated_entry.get('pid', None)),
            group=new_entry.group,
        )

    def delete_entry(self, entry_id: str) -> None:
        removed_id = self.delete_time_entry(entry_id)

        for i, entry in enumerate(self.entries):
            if entry.id == str(removed_id):
                del self.entries[i]
                break

    def push_session(self, session: dict, name: str, entry_id: str = '', project_id: str = None):
        headers = {
            "Content-Type": "application/json"
        }
        data = {
            'time_entry': {
                "description": name,
                "start": session['start_time'].isoformat() + "+00:00",
                "duration": int((session['end_time'] - session['start_time']).total_seconds()),
                "created_with": "thyme-toggl-cli"
            }
        }
        if project_id:
            data['time_entry']['pid'] = project_id
        if entry_id:
            return self.update_time_entry(entry_id, data)

        response = self.provider.request(
            'time_entries', data=json.dumps(data), headers=headers,
        )
        print(u'Pushed session to toggl: {}'.format(response))
        entry = response['data']
        entry['start_time'] = parse_time(entry['start'])
        entry['end_time'] = parse_time(entry['stop'])
        return entry

    def update_time_entry(self, entry_id: str, data: dict):
        response = self.provider.request(
            'time_entries/{}'.format(entry_id), data=json.dumps(data), method='PUT'
        )
        print(u'Updated session to toggl: {}'.format(response))
        entry = response['data']
        entry['start_time'] = parse_time(entry['start'])
        entry['end_time'] = parse_time(entry['stop'])
        return entry

    def delete_time_entry(self, entry_id):
        logger.info('deleting %s', entry_id)
        response = self.provider.request(
            'time_entries/{}'.format(entry_id), method='DELETE'
        )
        return response


class Provider(AbstractProvider):
    def __init__(self, api_key):
        self.api_key = api_key
        self.id_counter = 4

    def request(self, endpoint: str, **kwargs) -> Union[List[dict], dict]:
        url = 'https://www.toggl.com/api/v8/{}'.format(endpoint)
        kwargs['headers'] = kwargs.get('headers', {
            "Content-Type": "application/json"
        })
        kwargs['auth'] = kwargs.get('auth', (self.api_key, 'api_token'))

        method = kwargs.get('method', 'POST').lower()
        try:
            del kwargs['method']
        except KeyError:
            pass
        response = getattr(requests, method)(url, **kwargs)
        try:
            return response.json()
        except Exception as e:
            logger.exception("%s: %s", response.content, e)
            raise

    def test_request(self, endpoint: str, **kwargs) -> Union[List[dict], dict, str]:
        method = kwargs.get('method', 'POST').lower()
        if endpoint == "time_entries" and method == 'get':
            return [
                {
                    "id": "1",
                    "pid": "10",
                    "start": "2019-05-09T08:00:00+00:00",
                    "stop": "2019-05-09T09:00:00+00:00",
                    "description": "Toggl entry 1",
                },
                {
                    "id": "2",
                    "pid": "11",
                    "start": "2019-05-13T07:42:55+00:00",
                    "stop": "2019-05-13T08:34:52+00:00",
                    "description": "Toggl entry 2",
                },
                {
                    "id": "3",
                    "pid": "20",
                    "start": "2019-05-13T09:35:11+00:00",
                    "stop": "2019-05-13T10:34:02+00:00",
                    "description": "Toggl entry 3",
                }
            ]
        elif endpoint == "clients" and method == 'get':
            return [
                {
                    "id": "1",
                    "name": "First Client",
                },
                {
                    "id": "2",
                    "name": "Second Client",
                },
            ]
        elif endpoint.startswith("clients") and method == 'get':
            _clid = endpoint[8]
            return [
                {
                    "id": str(int(_clid) * 10),
                    "name": "Development"
                },
                {
                    "id": str(int(_clid) * 10 + 1),
                    "name": "Bug fixing"
                },
            ]
        elif endpoint == "time_entries" and method == 'post':
            entry = json.loads(kwargs['data'])['time_entry']
            entry['stop'] = (
                parse_time(entry['start']) + timedelta(seconds=entry['duration'])
            ).isoformat() + "+00:00"
            entry['id'] = self.id_counter
            self.id_counter += 1
            return {'data': entry}
        elif endpoint.startswith("time_entries") and method == 'put':
            entry = json.loads(kwargs['data'])['time_entry']
            entry['stop'] = (
                parse_time(entry['start']) + timedelta(seconds=entry['duration'])
            ).isoformat() + "+00:00"
            entry['id'] = endpoint[13:]
            return {'data': entry}
        elif endpoint.startswith("time_entries") and method == 'delete':
            return endpoint[13:]
        return [{}]



import requests
import json
from typing import List, cast, Any

from tracklater import settings
from .interfaces import IssueMixin, AbstractParser, AbstractProvider
from tracklater.models import Issue

import logging

logger = logging.getLogger(__name__)

API_URL = "https://api.github.com/graphql"


class Parser(IssueMixin, AbstractParser):
    def get_issues(self) -> List[Issue]:
        self.github_login()
        issues: List[Issue] = []
        for github_project in self.github_projects:
            for issue in self.provider.get_issues(
                github_project["repo"], github_project["id"]
            ):
                issues.append(
                    Issue(
                        key="#{}".format(issue["ref"]),
                        id=issue["id"],
                        title=issue["subject"],
                        group=github_project["group"],
                    )
                )
        return issues

    def github_login(self) -> None:
        github_settings = cast(Any, settings.GITHUB)
        self.provider = Provider(github_settings["global"]["TOKEN"])
        self.github_projects: List[dict] = []
        id_counter = 0
        for group, data in github_settings.items():
            if "repo" not in data:
                continue
            self.github_projects.append(
                {"id": id_counter, "group": group, "repo": data["repo"]}
            )
            id_counter += 1


class Provider(AbstractProvider):
    def __init__(self, token):
        self.token = token
        self.headers = {"Authorization": "Bearer {}".format(self.token)}

    def get_issues(self, repo, project_id):
        query = """
        query {
            repository(owner:"#{repo_owner}", name:"#{repo_name}") {
                issues(first:100#{cursor}) {
                    totalCount
                    edges {
                        node {
                            title
                            number
                            id
                        }
                    }
                    pageInfo {
                        endCursor
                        hasNextPage
                    }
                }
            }
        }
        """
        issues: List[dict] = []
        cursor = ""
        while True:
            logger.error(
                json.dumps(
                    {
                        "query": query.replace("#{repo_owner}", repo[0])
                        .replace("#{repo_name}", repo[1])
                        .replace("#{cursor}", cursor)
                    }
                )
            )
            response = requests.post(
                API_URL,
                data=json.dumps(
                    {
                        "query": query.replace("#{repo_owner}", repo[0])
                        .replace("#{repo_name}", repo[1])
                        .replace("#{cursor}", cursor)
                    }
                ),
                headers=self.headers,
            )
            logger.error(response.json())
            data = response.json()["data"]
            issues += [
                {
                    "ref": edge["node"]["number"],
                    "subject": edge["node"]["title"],
                    "id": edge["node"]["id"],
                }
                for edge in data["repository"]["issues"]["edges"]
            ]
            # Paginate
            if data["repository"]["issues"]["pageInfo"]["hasNextPage"]:
                cursor = ' after:"{}"'.format(
                    data["repository"]["issues"]["pageInfo"]["endCursor"]
                )
            else:
                break
        return issues
